#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head1 NAME

DBI -- Database Interface for WDB.

=head1 Description

This class defines the database interface for WDB. Interfaces to 
specific database systems iherit from this class.


=head1 Public Methods

=cut #--------------------------------------------------------------------------

package WDB::DBI;

use English;                    	# Use long "english" variable names 
$debug = $main::debug;
$IC_UPPER = 1;   # TEST: Ignore Case using UPPER()

#ERROR CODES
$ERR_UNIQ    = 'ERR_UNIQ';
$ERR_UNKNOWN = 'UNKNOWN';

my(%DBI_CACHE);
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 factory( $system, $user, $pswd, $server, $database )

New creates a new DBI object for a given database system (DBI sub-class).

Returns: DBI object
=cut
#------------------------------------------------------------------------------
sub factory 
{
    my($system, $user, $pswd, $server, $database) = @_;
    my($key) = join(':',@_);
    my($dbi);

    if ( @{$DBI_CACHE{$key}} ) {
	$dbi = pop @{$DBI_CACHE{$key}};
    }

    require 'WDB/DBI_'.$system.'.pm'
	|| WDB::Error::Fatal("Error loading database interface file WDB/DBI_$system.pm :\n$!\n");
    $dbi = eval("new WDB::DBI_$system('$user','$pswd','$server','$database')");
    if ( $@ ) { WDB::Error::Fatal("new WDB::DBI_$system failed: $@\n") };

    return $dbi;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 cache()

Caches the current DBI object, so the connection can be re-used.


=cut 
#------------------------------------------------------------------------------
sub cache 
{
    my($self) = shift;
    my($key);

    my($system, $user, $pswd, $server, $database) = @_;
    
    $key = $self->{'system'}.':'.$self->{'user'}.':'.$self->{'pswd'}.':'.
	   $self->{'server'}.':'.$self->{'database'};

    print main::DEBUG "WDB::DBI::cache - Caching $key\n" if $debug;

    push( @{$DBI_CACHE{$key}}, $self );

}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 new( $user, $pswd, $server, $database )

New creates a new database connection object.

=cut
#------------------------------------------------------------------------------
sub new 
{
    my($type) = shift;
    my($self) = {};
    my($user, $pswd, $server, $database) = @_;
    bless ($self, $type);

    $self->{user}     = $user;
    $self->{pswd}     = $pswd;
    $self->{server}   = $server;
    $self->{database} = $database;

    return $self;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 do( sql_statement )

Executes the given SQL statement. This is for I<one-shot> SQL statements.
For selects use the open/fetch/close methods.

=cut
#------------------------------------------------------------------------------
sub do
{
    my($self) = shift;
    my($sql)  = @_;

    WDB::Error::Fatal("WDB::DBI::do Not Implemented!\n");
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 try( sql_statement )

Same as do, but doesn't die on error. Instead true or false is returned.

=cut
#------------------------------------------------------------------------------
sub try
{
    my($self) = shift;
    my($sql)  = @_;

    WDB::Error::Fatal("WDB::DBI::try Not Implemented!\n");
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 getLastError()

Returns last error. 

Usage:
    ($err_code,$native_err_num, $native_err_msg) = $dbi->getLastError();

=cut
#------------------------------------------------------------------------------
sub getLastError
{
    my($self) = shift;

    return(0, 0, 'Unknown Error.'); 
}

 
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 open( sql_statement )

Prepares and executes the given SQL select statement.
The returned rows can be retrieved via the nextrow method.

I<In the old WDB dbi this used to be called dbi_dosql, but open seemed like
  a more appropriate name...>

=cut
#------------------------------------------------------------------------------
sub open
{
    my($self) = shift;
    my($sql)  = @_;

    WDB::Error::Fatal("WDB::DBI::open Not Implemented!\n");
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 close()

Closes a previously open'ed statement.

=cut
#------------------------------------------------------------------------------
sub close
{
    my($self) = shift;

    WDB::Error::Fatal("WDB::DBI::close Not Implemented!\n");
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 nextrow()

Gets the next row from a previous select (dbi_dosql).

After a open("select ... ") call, this function can be 
called repeatetly to retrieve all rows of the query.

Example :
	$dbi->open("select * from mytab");
	while( %row = $dbi->nextrow ) {
	    print $row{'columnname'};
	    ...
	}

Returns	an associative array (keyed on the column name) of formatted 
data, based on the datatype of the corresponding columns.

=cut
#------------------------------------------------------------------------------
sub nextrow
{
    my($self) = shift;

    WDB::Error::Fatal("WDB::DBI::nextrow Not Implemented!\n");
}
 


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 disconnect()

Disconnects the current database connection

After this function is called there are no current database 
connection => no other function from this package can be 
called. 

This method is automatically called when the current object is
destroyed.

=cut
#------------------------------------------------------------------------------
sub disconnect
{
    my($self) = shift;

    warn "WDB::DBI::disconnect Not Implemented!\n";
}
 
sub DESTROY
{
    my($self) = shift;

    $self->disconnect;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 commit

Commits SQL statements executed since last commit

=cut
#------------------------------------------------------------------------------
sub commit
{
    my($self) = shift;

    WDB::Error::Fatal("WDB::DBI::commit Not Implemented!\n");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 rollback

Rollsback SQL statements executed since last commit

=cut
#------------------------------------------------------------------------------
sub rollback
{
    my($self) = shift;

    WDB::Error::Fatal("WDB::DBI::rollback Not Implemented!\n");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 autocommit( on_or_off )

Enables automatic commit after each executed SQL statement.

If I<on_or_off> is true (non-zero) autocommit is enabled. If it is
false (zero) is is disabled.

=cut
#------------------------------------------------------------------------------
sub autocommit
{
    my($self) = shift;

    WDB::Error::Fatal("WDB::DBI::autocommit Not Implemented!\n");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 rowcount

Sets the maximim number of rows returned from a query.

Causes SQL server to stop processing the query ( select, 
insert update, delete) after the specified number of rows 
are affected.

NOTE: This function is added for efficiency only. 
When used with WDB it can safely be ignored.

=cut
#------------------------------------------------------------------------------
sub rowcount
{
    my($self) = shift;
}
 
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 dateformat( format )

Sets the default output format for dates.

Sets the order of the date parts month/day/year for entering
datetime data. Valid arguments are mdy, dmy, ymd, ydm, myd, dym.

=cut
#------------------------------------------------------------------------------
sub dateformat
{
    my($self)   = shift;
    my($format) = shift;
}


#==============================================================================
#==============================================================================
=head1 High Level Methods

Following are a set of High Level Functions which utilize the above 
I<Low level functions> to provide a more general database interface, which in 
theory is independant of the specific SQL version used by a given database 
system.

When writing a DBI for a database system with a fairly standard SQL you
will probably not have to change any of these methods, but can simply
inherit them from WDB::DBI.

At the moment this is far from complete, but eventually WDB will only reference
these methods, making it a lot easier to port WDB to different database 
systems. This will also allow for better utilization of various database
system specific optimizations, and even for non SQL database systems.


=cut#==========================================================================
#==============================================================================


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 getnextserialkey( table )

Gets the next serial key value for a particular table.

Requires a table named serial_key, and an entry in that table
for each table that has a serial key value.  The value in the
table is retrieved, incremented by one, and the new value is
inserted into the serial_key table.

    Table : serial_key
    Fields: table_name, key_value

create table serial_key (
	table_name	char(30),
	key_value	int
	);

=cut
#------------------------------------------------------------------------------
sub getnextserialkey
{
    my($self)  = shift;
    my($table) = shift;
    my($sql, %row);

    $self->autocommit( 0 );

    $self->do( "update serial_key set key_value = key_value+1 "
	     . "where table_name = '$table'");

    $self->open( "select key_value KEY_VALUE from serial_key "
	       . "where table_name = '$table'");

    if ( ! (%row = $self->nextrow() ) ) {
	WDB::Error::Fatal("Error: Couldn't get next key for $table\n");
    }


    $self->commit( $dbproc );
    $self->autocommit( 1 );

    return $row{'KEY_VALUE'};
}

sub use_fieldnames
{
    my($self)  = shift;
    return 0;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 load_next_row

Fetches the next row from the current query and loads the values into
the fields of the given FDF.

TODO: This should probably be changed in next version.

=cut
#------------------------------------------------------------------------------
sub load_next_row
{
    my($self)  = shift;
    my($fdf) = @_;
    my(%row);
    my($val,$name);
    my(@comp_fields);
    my($f);

#   if ( %row = $self->nextrow ) {
#	foreach $key ( keys( %row ) ) {
#	    if ( defined  $fdf->{field}{$key} ) {
#		$fdf->{field}{$key}->{value} = $row{$key};
#	    } else {
#		die "Error in DBI::load_next_row"
#		  . " - Field $key not defined in FDF\n";
#	    }
#	}
#   }

    if ( %row = $self->nextrow ) {
	$self->load_row($fdf,%row);

#	foreach $f ( @{$fdf->{fields}} ) {
#	    if ( defined $f->{computed} ) {	# Do computed fields last.
#		push( @comp_fields, $f );
#		next;
#	    }
#	    $f->set_db_value( $row{$f->{FIELD}} );
#	}
#        foreach $f ( @comp_fields ) {
#	    $f->set_db_value( $row{$f->{FIELD}} );
#	}
    }

    return %row;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 load_row

loads the values in the given row hash into the fields of the given FDF.


=cut
#------------------------------------------------------------------------------
sub load_row
{
    my($self)  = shift;
    my($fdf, %row) = @_;
    my($val,$name);
    my(@comp_fields);
    my($f);

	foreach $f ( @{$fdf->{fields}} ) {
	    if ( defined $f->{computed} ) {	# Do computed fields last.
		push( @comp_fields, $f );
		next;
	    }
	    $f->set_db_value( $row{$f->{FIELD}} );
	}
        foreach $f ( @comp_fields ) {
	    $f->set_db_value( $row{$f->{FIELD}} );
	}
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 update( cgi, fdf )

Updates a singel row in the database from the given user input in cgi and 
the field definitions in the fdf.

TODO: This is another  quick-and-dirty....

=cut
#------------------------------------------------------------------------------
sub update
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($table,$val, @where,$data);
    my(@update_fields,@update_values,@update_list);


    ($table) = $fdf->{'TABLE'} =~ /([^,]+).*/; 

    #-------------------------------------------------------------
    # Compose update fields and set list for the SQL statement.
    #-------------------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {
	print main::DEBUG "Update: $f->{FIELD}\n" if $debug;

	#------------------------
	# Key field ?
	#------------------------
        if ( defined $f->{'key'} ) {
	    print main::DEBUG "Update: $f->{FIELD} - key" if $debug;

	    $f->{oper} = '=='; # Allways use exact match on keys

            if ( defined $f->{'serial_key'} ) {
	 	print main::DEBUG "serial_key" if $debug;
                $serial = 1;
            }
	    $val = $f->get_db_value();
            if ( ! defined $val || $val eq '' ) {
                WDB::Error::Fatal("Fatal error: missing key data. $f->{FIELD}");
            } else {
		#push( @where, eval('$self->query_' . $f->db_type 
		#	       . '($f->{column},$f->{oper},$f->{value})') );
	        push( @where, eval('$self->query_' . $f->db_type 
			       . '($f->{column},$f->{oper},$f->get_db_value())') );
            }
	    print main::DEBUG "\n" if $debug;

	#------------------------
	# Update field
	#------------------------
        } else {
	    $f->{value} =~ s/^\s*//;	# Remove leading and trailing blanks
	    $f->{value} =~ s/\s*$//;
	    next if defined $f->{ 'foreign_key'};
	    next if defined $f->{ 'no_update'};

            # Don't update fields that wheren't in the form....
	    #next if ! defined $cgi->param($f->{FIELD}); 
	    #next if ! defined $f->{value}; 
	    next if $f->{oper} ne '='; 

	    if ( defined $f->{ 'computed'} ) {
	        print main::DEBUG "    - : Computed field\n" if $debug;
		# Get value - just to do the eval stuff...
	        $data = $f->get_db_value;
		next;
	    }

	    print main::DEBUG "    - : -\n" if $debug;
            #---------------------------------------------------------
            # NULL 'char' fields need to be compared using 'ne ""'.
            # NULL numerical fields need to be compared to 0.
            #---------------------------------------------------------
            if ( defined $f->{value}
                 && (($f->{ 'type'} eq "char" && $f->{value} ne "")
                      || ($f->{ 'type'} ne "char" && $f->{value} ne "")) ) {
		# CHANGED!! Did we realy need this ? Maybe I should add a 
		# null_on_zero attribute if this is needed ?
                #      || ($f->{ 'type'} ne "char" && $f->{value} != 0)) ) {
		push( @update_fields, ( $f->{'column'} )
		    ?  $f->{'column'} : $f->{FIELD} );
		print main::DEBUG "$f->{FIELD} added to update_fields\n" if $debug;
		print main::DEBUG "update_fields = " . join(',',@update_fields)."\n" if $debug;
	        $data = $f->get_db_value;
		if ( $f->{ 'type'} eq "char" ) {
		    #----------------------------------------
		    # Escape imbedded apostrophes and quotes.
		    #----------------------------------------
		    #$data =~ s/'/\\'/g;
		    #$data =~ s/"/\\"/g;
		    $data = "'".$data."'";
		    push( @update_values, $data );
		    print main::DEBUG "$f->{FIELD} value = $data\n" if $debug;
		} elsif ( $f->{ 'type'} eq "date" and $data ne 'SYSDATE' ) {
		    $data = "'".$data."'";
		    push( @update_values, $data );
		} else {
		    print main::DEBUG "$f->{FIELD} value = $data\n" if $debug;
		    push( @update_values, "$data" );
		}
            } else {
		push( @update_fields, ( $f->{'column'} )
		    ?  $f->{'column'} : $f->{FIELD} );
		print main::DEBUG "$f->{FIELD} added to update_fields as NULL\n" if $debug;

                ###
		push( @update_values, 'NULL' );
		#if ( $f->{ 'type'} eq "char" ) {
		#    $data = "''";
		#    push( @update_values, $data );
		#} else {
		#    push( @update_values, 0 );
		#}
            }
        }
    }
    #-------------------------------
    # Update the database
    #-------------------------------

    print main::DEBUG "Updating singel table (no JOIN)\n" if $debug;
    $update = "update $table set ";

    for $f (@update_fields) {
	$update_temp = $f."=".shift(@update_values);
	print main::DEBUG "Updating $f : $update_temp\n" if $debug;
	push (@update_list, $update_temp);
	undef $update_temp;
    }
    $update_list = join(',', @update_list);

    #$self->do( "$update $update_list where " . join(' AND ',@where) );
    if ( $self->try( "$update $update_list where " . join(' AND ',@where)) ) {
    } else {
	return 0;
    }
    return 1;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 insert( cgi, fdf )

Inserts a singel row in the database from the given user input in cgi and 
the field definitions in the fdf.

TODO: This is yet another quick-and-dirty....

Returns: 1 on success, 0 on failure. (use getLastError() to get error info.)

=cut
#------------------------------------------------------------------------------
sub insert
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($table, @tables, $data, $insert_cmd);


    #-------------------------------------------------------------
    # Always insert in the first table in the TABLE FDF attribute
    #-------------------------------------------------------------

    if ( defined $fdf->{'INSERT'} ) {
	@tables = split(',',$fdf->{'INSERT'} );
    } else {
	WDB::Error::Fatal("Insert not allowed in $fdf->{'NAME'}. INSERT attribute not set."); 
    }

    #-------------------------------------------------------------
    # Insert data in each table.
    #-------------------------------------------------------------
    foreach $table ( @tables ) {
	my( @insert_fields, @insert_values );

	#-------------------------------------------------------------
	# Compose insert fields and value list for the SQL statement.
	#-------------------------------------------------------------
	foreach $f ( @{$fdf->{fields}} ) {
	    if ( defined $f->{'computed'} ) {
		$data = $f->get_db_value; # Just do the computation
		next;			  # - then skip to the next field.
	    }

	    my( $tab, $col ) = split('\.', $f->{'column'}, 2 );
	    print main::DEBUG "Insert tab=$tab, col=$col\n" if $debug;
	    next if ( $tab ne $table );

	    print main::DEBUG "Insert: $f->{FIELD}\n" if $debug;

	    #------------------------
	    # Serial key field ?
	    #------------------------
	    if ( defined $f->{'serial_key'} ) {
		print main::DEBUG "serial_key\n" if $debug;
		$f->set_db_value( $self->getnextserialkey($table) );
	    }


	    #------------------------------
	    # Should we skip this field ?
	    #------------------------------

	    #next if defined $f->{ 'foreign_key'};
	    next if defined $f->{ 'no_insert'};
	    #next if defined $f->{ 'hidden'};

	    #---------------------------------------------------------
	    # Compose field and value list
	    #---------------------------------------------------------

	    push( @insert_fields, $col );
	    $data = $f->get_db_value;
	    if ( $f->{ 'type'} eq 'char' ) {
		$data = "'".$data."'";
		push( @insert_values, $data );
	    } elsif ( $f->{ 'type'} eq 'date' && $data eq 'SYSDATE' ) {
		push( @insert_values, $data );
	    } elsif ( $f->{ 'type'} eq 'date' && $data ne 'NULL' ) {
		$data = "'".$data."'";
		push( @insert_values, $data );
	    } else {
		push( @insert_values, $data );
	    }
	}

	#-------------------------------
	# Insert row into the database
	#-------------------------------

	$insert_cmd  = "insert into $table ("
		     . join(',', @insert_fields)
		     . ") values ("
		     . join(',', @insert_values)
		     . ")";

	#print $insert_cmd;
	if ($self->try( $insert_cmd )) {
	} else {
	    return 0;
	}
    } 
    return 1;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 delete( cgi, fdf )

delete a singel row from the database from the given user input in cgi and 
the field definitions in the fdf.

TODO: This is another yet quick-and-dirty....
NEW - added support for multiple fields.

=cut
#------------------------------------------------------------------------------
sub delete
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($table);
    my(@wherelist);
    my($where);

    ($table) = $fdf->{'TABLE'} =~ /([^,]+).*/; 


    #-------------------------------------------------------------
    # Compose update fields and set list for the SQL statement.
    #-------------------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {
	print main::DEBUG "Delete: $f->{FIELD}\n" if $debug;

	#------------------------
	# Key field ?
	#------------------------
        if ( defined $f->{'key'} ) {
	    print main::DEBUG "Delete: $f->{FIELD} - key" if $debug;

            if ( ! defined $main::in{$f->{FIELD}} ) {
                WDB::Error::Fatal("Fatal error: missing key data.");
            } else {
		#push( @wherelist, eval('$self->query_' . $f->db_type 
		#	       . '($f->{column},"==",$main::in{$f->{FIELD}})') );
	        push( @wherelist, eval('$self->query_' . $f->db_type 
			       . '($f->{column},"==",$f->get_db_value())') );
            }
        }
    }
    $where = "where " . join(' AND ', @wherelist );

    #-------------------------------
    # Delete the row
    #-------------------------------
    print main::DEBUG "DELETE FROM $table $where" if $debug;
    $self->do( "delete from $table $where");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 change( cgi, fdf )

Updates many rows in the database from the given user input in cgi and 
the field definitions in the fdf.

Each field can be updated either by replacment, +/- a value or a percentage 

=cut
#------------------------------------------------------------------------------
sub change
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($table,$val,$err,@where,$data);
    my(@update_fields,@update_values,@update_list);

    my($f,$col);

    #--------------------------------------------------
    # Compose select and where list for the SQL select
    #--------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {
	print main::DEBUG "Processing $f->{FIELD} ...\n" if $debug;

	$col = $f->{'column'};

	#--------------------------------------------------
	# Compose query on each field.
	#--------------------------------------------------
	if ( defined $f->{value} && $f->{value} ne "" ) {
	    next if $f->{value} eq '%';    # Ignore '%' fields (from enum fields)
	    print main::DEBUG "query on $f->{FIELD} $f->{oper} $f->{value} \n" if $debug;

	    #--------------------------------------------------
	    # Computed fields
	    #--------------------------------------------------
	    if ( defined $f->{'computed'} ) { 
	      print main::DEBUG "Evaluating computed field : $f->{FIELD} = $f->{value}}\n" if $debug;
	      print main::DEBUG "- Compute : $f->{'to_db'}\n" if $debug;

	      $ret = WDB::Eval::eval("$f->{FIELD}:to_db",$f->{to_db});
	      print main::DEBUG " = $ret\n" if $debug;
	      next;	# Don't query on computed fields.
	    }

	    push( @where, eval('$self->query_' . $f->db_type 
			       . '($f->{column},$f->{oper},$f->get_db_value())') );

	}
    }
    # TODO: Different CONSTRAINTS on change ???
    push( @where, $fdf->{'CONSTRAINTS'}) if $fdf->{'CONSTRAINTS'};


    ($table) = $fdf->{'TABLE'} =~ /([^,]+).*/; 

    #-------------------------------------------------------------
    # Compose update fields and set list for the SQL statement.
    #-------------------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {

	$val = $cgi->param('change::'.$f->{FIELD});
	$val =~ s/^\s*//;	# Remove leading and trailing blanks
	$val =~ s/\s*$//;
	next if ( $val eq '' );

	# TODO: Verify ?
	#$f->{value} = $val;
	$f->{not_null} = 0;
	if ( $f->verify($val) ) { 
	    $err++;
	} else {

	    $col = ( $f->{'column'}  ?  $f->{'column'} : $f->{FIELD} );

	    if ( $f->{ 'type'} eq "char" || $f->{ 'type'} eq "date" ) {
		$data = $f->get_db_value;
		$data = "'".$data."'";
		push (@update_list, $col."=".$data);
	    } else {
		$oper = $cgi->param('change_type::'.$f->{FIELD});
		if ( $oper eq 'plus' ) {
		    $data = $f->get_db_value;
		    push (@update_list, "$col = $col + $data");
		} elsif ( $oper eq 'pct' ) {
		    $data = $f->{value};
		    push (@update_list, "$col = $col * ((100+$data)/100)");
		} else { # default: replace
		    $data = $f->get_db_value;
		    push (@update_list, $col."=".$data);
		}
	    }
	}
    }
    return 0 if $err;

    #-------------------------------
    # Update the database
    #-------------------------------

    $update = "update $table set ";

    if ( $#update_list >= 0 ) {
	$update_list = join(',', @update_list);

	if ( $self->try( "$update $update_list where " . join(' AND ',@where)) ) {
	} else {
	    return 0;
	}
    }
    return 1;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 prepare_select( cgi, fdf )

Prepares a select statement from the given user input in cgi and 
the field deffinitions in the fdf.

The conversion from WDB
types (user types) to database types are done in the WDB::Field_xxx
objects, and the generation of query strings given the database type
and operator are done in speical DBI methods per type.
This allows for overwriting just the needed bit of behavoiur when 
porting the DBI to a new database system, and also allows for an
easy way to add new WDB data types.

TODO: The following is not supported yet :
      o  enum + multiple 
      o  OR queries
      o  Range queries (..)


=cut
#------------------------------------------------------------------------------
sub prepare_select
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($f, $col, @where);

    #--------------------------------------------------
    # Compose select and where list for the SQL select
    #--------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {
	print main::DEBUG "Processing $f->{FIELD} ...\n" if $debug;

	if ( ! defined $f->{'computed'} ) {
	    if ( $self->use_fieldnames ) {
		push( @select, $f->{'column'} ); 
		push( @select_fields, "$f->{FIELD}" );
	    } else {
		push( @select, "$f->{'column'} $f->{FIELD}" );
	    }
	}

	$col = $f->{'column'};

	#--------------------------------------------------
	# Compose query on each field.
	#--------------------------------------------------
	if ( defined $f->{value} && $f->{value} ne "" ) {
	    next if $f->{value} eq '%';    # Ignore '%' fields (from enum fields)
	    print main::DEBUG "query on $f->{FIELD} $f->{oper} $f->{value} \n" if $debug;

	    #--------------------------------------------------
	    # Computed fields
	    #--------------------------------------------------
	    if ( defined $f->{'computed'} ) { 
	      print main::DEBUG "Evaluating computed field : $f->{FIELD} = $f->{value}}\n" if $debug;
	      print main::DEBUG "- Compute : $f->{'to_db'}\n" if $debug;

	      $ret = WDB::Eval::eval("$f->{FIELD}:to_db",$f->{to_db});
	      print main::DEBUG " = $ret\n" if $debug;
	      next;	# Don't query on computed fields.
	    }

	    push( @where, eval('$self->query_' . $f->db_type 
			       . '($f->{column},$f->{oper},$f->get_db_value())') );

	}
    }
    push( @where, $fdf->{'CONSTRAINTS'}) if $fdf->{'CONSTRAINTS'};

    push( @where, $fdf->{'JOIN'}) if $fdf->{'JOIN'};

    $selectlist = join(',',@select);

    $wherelist  = join(' AND ', @where );
    $wherelist  = "where ".$wherelist if $wherelist;

    $wherelist .= " order by $fdf->{'ORDER'}" if $fdf->{'ORDER'};


    #------------------------
    # EXECUTE QUERY
    #------------------------


    $self->fieldnames( @select_fields ) if $self->use_fieldnames();

    if ( $debug ) {
	print main::DEBUG "\n**** QUERY ****\n";
	print main::DEBUG "use $fdf->{'DATABASE'}\n";
	print main::DEBUG "select $selectlist from $fdf->{'TABLE'}\n";
	print main::DEBUG "$wherelist \n\n";
    }
    
    $self->open( "select $selectlist from $fdf->{'TABLE'} $wherelist");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 query_char($column,$oper,$value) 

Returns the SQL to query on a value using the given operator in a char column.
F.ex. 
    query_char('tab.col','=','test') ==>  "tab.col like '%test%'"

=cut
#------------------------------------------------------------------------------
sub query_char
{
    my($self) = shift;
    my($col,$oper,$val) = @_;

    if ( $oper ne '==' ) {
	$val =~ s/^\s*//;
	$val =~ s/\s*$//;
    }
    print main::DEBUG "      val = '$val'\n" if $debug;

    $val  =~ s/"/""/g;
    $val  =~ s/'/''/g;
    if ( $oper eq '~~' ) {
	$oper = "like";
	$val  = "%$val%";
	if ( $IC_UPPER ) {
	   $col = "NLS_UPPER( $col, 'NLS_SORT = XDanish')";
	   #$val  =~ s/([a-z])/\U$1/g;
	   $val  = "NLS_UPPER('$val', 'NLS_SORT = XDanish')";
	} else {
	   $val  =~ s/([a-zA-Z])/[\U$1\L$1]/g;
	   $val  = "'$val'";
	}
    } elsif ( $oper eq "=" ) {
	$oper = "like";
	$val  = "'$val'";
    } elsif ( $oper eq "==" ) {
	$oper = "=";
	$val  = "'$val'";
    } elsif ( $oper eq "~" ) {
	$oper = "like";
	if ( $IC_UPPER ) {
	   $col = "NLS_UPPER( $col, 'NLS_SORT = XDanish' )";
	   #$val  =~ s/([a-z])/\U$1/g;
	   $val  = "NLS_UPPER('$val', 'NLS_SORT = XDanish')";
	} else {
	   $val  =~ s/([a-zA-Z])/[\U$1\L$1]/g;
	   $val  = "'$val'";
	}
    }
    $val =~ s/\*/%/g if $oper eq "like";
    $val =~ s/\?/_/g if $oper eq "like";

    return( "$col $oper $val" );
}



#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 query_datetime($column,$oper,$value) 

Returns the SQL to query on a value using the given operator in a datetime column.
F.ex. 
    query_datetime('tab.col','=','1-jan-97') ==>  "tab.col like '1-jan-97'"

=cut
#------------------------------------------------------------------------------
sub query_datetime
{
    my($self) = shift;
    my($col,$oper,$val) = @_;

    $val =~ s/^\s*//;
    $val =~ s/\s*$//;

    print main::DEBUG "      val = '$val'\n" if $debug;

    if ($oper eq "==" ) {
	$oper = "=" 
    } else {
	$oper = "like" unless $oper;    # Defaul operand is 'like'
	$val =~ s/\*/%/g;
    }

    return( "$col $oper $val" ) if ( $val eq 'NULL' or $val eq 'SYSDATE' ) ;
    return( "$col $oper '$val'" );
}



#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 query_float($column,$oper,$value) 

Returns the SQL to query on a value using the given operator in a float column.
F.ex. 
    query_float('tab.col','>=','3.1415') ==>  "tab.col >= 3.1415"

=cut
#------------------------------------------------------------------------------
sub query_float
{
    my($self) = shift;
    my($col,$oper,$val) = @_;

    $oper = "=" if $oper eq "==";
    return( "$col $oper $val");
}



#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 query_int($column,$oper,$value) 

Returns the SQL to query on a value using the given operator in an int column.
F.ex. 
    query_int('tab.col','>=','42') ==>  "tab.col >= 42"

=cut
#------------------------------------------------------------------------------
sub query_int
{
    my($self) = shift;
    my($col,$oper,$val) = @_;

    $oper = "=" if $oper eq "==";
    return( "$col $oper $val");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 update( cgi, fdf )

Updates a singel row in the database from the given user input in cgi and 
the field definitions in the fdf.

TODO: This is another  quick-and-dirty....

=cut
#------------------------------------------------------------------------------
sub __old__update
{
    my($self)  = shift;
    my($cgi, $fdf) = @_;
    my($table);


    ($table) = $fdf->{'TABLE'} =~ /([^,]+).*/; 

    #-------------------------------------------------------------
    # Get qualification either from the form or from the path
    #-------------------------------------------------------------
    if (  $cgi->param('wdb_keys') ) {
        print main::DEBUG "Query using keys\n" if $debug;
	@keys = split( '::', $cgi->param('wdb_keys') );
	foreach $key ( @{$fdf->{keys}} ) {
	    $main::in{$key} = shift @keys;
            print main::DEBUG " - $key = $main::in{$key}\n" if $debug;
	}
    }

    #-------------------------------------------------------------
    # Compose update fields and set list for the SQL statement.
    #-------------------------------------------------------------
    foreach $f ( @{$fdf->{fields}} ) {
	print main::DEBUG "Update: $f->{FIELD}\n" if $debug;

	#------------------------
	# Key field ?
	#------------------------
        if ( defined $f->{'key'} ) {
	    print main::DEBUG "Update: $f->{FIELD} - key" if $debug;

            if ( defined $f->{'serial_key'} ) {
	 	print main::DEBUG "serial_key" if $debug;
                $serial = 1;
            }
            if ( ! defined $main::in{$f->{FIELD}} ) {
                WDB::Error::Fatal("Fatal error: missing key data.");
            } else {
		# TODO: Add support for multiple keys
                $serial_key = $main::in{$f->{FIELD}};
                if ( $f->{ 'type'} eq "char" ) {
                    $where = " where ".$f->{'column'}."='".$serial_key."'";
                } else {
                    $where = " where ".$f->{'column'}."=".$serial_key;
                }
	 	print main::DEBUG "WHERE= $where" if $debug;
            }
	    print main::DEBUG "\n" if $debug;

	#------------------------
	# Update field
	#------------------------
        } else {
	    $main::in{$f->{FIELD}} =~ s/^\s*//;	# Remove leading and trailing blanks
	    $main::in{$f->{FIELD}} =~ s/\s*$//;
	    next if defined $f->{ 'foreign_key'};
	    next if defined $f->{ 'computed'};
	    next if defined $f->{ 'no_update'};
	    next if defined $f->{ 'hidden'};

            # Don't update fields that wheren't in the form....
	    next if ! defined $cgi->param($f->{FIELD}); 

            #---------------------------------------------------------
            # NULL 'char' fields need to be compared using 'ne ""'.
            # NULL numerical fields need to be compared to 0.
            #---------------------------------------------------------
            if ( defined $main::in{$f->{FIELD}}
                 && (($f->{ 'type'} eq "char" && $main::in{$f->{FIELD}} ne "")
                      || ($f->{ 'type'} ne "char" && $main::in{$f->{FIELD}} != 0)) ) {
		push( @update_fields, ( $f->{'column'} )
		    ?  $f->{'column'} : $f->{FIELD} );
		print main::DEBUG "$f->{FIELD} added to update_fields\n" if $debug;
		print main::DEBUG "update_fields = " . join(',',@update_fields)."\n" if $debug;
		#$data = $main::in{$f->{FIELD}};
	        $data = $f->get_db_value;
		if ( $f->{ 'type'} eq "char" ) {
		    #----------------------------------------
		    # Escape imbedded apostrophes and quotes.
		    #----------------------------------------
		    #$data =~ s/'/\\'/g;
		    #$data =~ s/"/\\"/g;
		    $data = "'".$data."'";
		    push( @update_values, $data );
		    print main::DEBUG "$f->{FIELD} value = $data\n" if $debug;
		} else {
		    print main::DEBUG "$f->{FIELD} value = $data\n" if $debug;
		    push( @update_values, "$data" );
		}
            } else {
		push( @update_fields, ( $f->{'column'} )
		    ?  $f->{'column'} : $f->{FIELD} );
		print main::DEBUG "$f->{FIELD} added to update_fields as NULL\n" if $debug;
		if ( $f->{ 'type'} eq "char" ) {
		    $data = "''";
		    push( @update_values, $data );
		} else {
		    push( @update_values, 0 );
		}
            }
        }
    }
    #-------------------------------
    # Update the database
    #-------------------------------

    print main::DEBUG "Updating singel table (no JOIN)\n" if $debug;
    $update = "update $table set ";

    for $f (@update_fields) {
	$update_temp = $f."=".shift(@update_values);
	print main::DEBUG "Updating $f : $update_temp\n" if $debug;
	push (@update_list, $update_temp);
	undef $update_temp;
    }
    $update_list = join(',', @update_list);

    #print "$update $update_list $where";
    $self->do( "$update $update_list $where");
}


1; #			    ------ EOF -----
