#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head1 NAME

DBI_ODBC -- Database Interface for ODBC.

=head1 Description

This class defines the database interface that WDB uses to access 
ODBC databases.

=head1 Public Methods
=cut#--------------------------------------------------------------------------

use WDB::DBI;

package WDB::DBI_ODBC;@ISA = qw( WDB::DBI );
@WDB::DBI_ODBC::Inherit::ISA = @ISA;	# Access to overridden methods.
use English;                    		# Use long "english" variable names 
use Win32::ODBC;

$debug = $main::debug;


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 new( user, pswd, server, database )

New logs in to a database and creates a new database connection object.

=cut
#------------------------------------------------------------------------------
sub new 
{
    my($type) = shift;
    my($self) = new WDB::DBI(@_); # First run the DBI constructor...
    bless $self, $type;		  # Then re-bless so we are a DBI_ODBC object.

	my($DSN) = 'DSN='.$self->{server}.';UID='.$self->{user}.';PWD='.$self->{pswd}; 

	if ( ! ($self->{dbproc} = new Win32::ODBC($DSN))) {
		WDB::Error::Fatal( "Error connecting to $DSN\n"
						  ."Error: " . Win32::ODBC::Error() . "\n");
	}

    return $self;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 do( sql_statement )

Executes the given SQL statement. This is for I<one-shot> SQL statements.
For selects use the open/fetch/close methods.

Dies on error !

=cut
#------------------------------------------------------------------------------
sub do
{
    my($self) = shift;
    my($sql)  = @_;

    print main::DEBUG "DBI_ODBC::do( $sql ) \n" if $main::debug;
    if ( ! $self->try($sql) ) {
	WDB::Error::Fatal("ODBC ERROR:<P>".Win32::ODBC::Error()."\n<P>Used SQL:<P>\n".$sql);
    }
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 try( sql_statement )

Same as do, but doesn't die on error. Instead true or false is returned.

=cut
#------------------------------------------------------------------------------
sub try
{
    my($self) = shift;
    my($sql)  = @_;

    print main::DEBUG "DBI_ODBC::try( $sql ) \n" if $main::debug;

	if ( $self->{dbproc}->Sql($sql)) {
                print main::DEBUG "DBI_ODBC::try - FAILED:".Win32::ODBC::Error()."\n" if $main::debug;
		return 0;
	}
	return 1;
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 getLastError()

Returns last error. 

Usage:
    ($err_code,$native_err_num, $native_err_msg) = $dbi->getLastError();

=cut
#------------------------------------------------------------------------------
sub getLastError
{
    my($self) = shift;
    my($ErrNum, $ErrText, $ErrConn) = $self->{dbproc}->Error();
    my($ErrCode) = $WDB::DBI::ERR_UNKNOWN;

    $ErrCode = $WDB::DBI::ERR_UNIQ if ( $ErrNum == 1 ) ;

    return($ErrCode, $ErrNum, $ErrText); 
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 open( sql_statement )

Prepares and executes the given SQL select statement.
The returned rows can be retrieved via the fetch method.

Dies on error !

=cut
#------------------------------------------------------------------------------
sub open
{
    my($self) = shift;
    my($sql)  = @_;

    print main::DEBUG "DBI_ODBC::open( $sql ) \n" if $main::debug;
    $self->do($sql);
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 close()

Closes a previously open'ed statement.

=cut
#------------------------------------------------------------------------------
sub close
{
    my($self) = shift;

    print main::DEBUG "DBI_ODBC::close() \n" if $main::debug;
    $self->{dbproc}->Close() # || WDB::Error::Fatal(Win32::ODBC::Error());
}
 

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 nextrow()

Gets the next row from a previous select (dbi_dosql).

After a open("select ... ") call, this function can be 
called repeatetly to retrieve all rows of the query.

Example :
	$dbi->open("select * from mytab");
	while( %row = $dbi->nextrow ) {
	    print $row{'columnname'};
	    ...
	}

Returns	an associative array (keyed on the column name) of formatted 
data, based on the datatype of the corresponding columns.

=cut
#------------------------------------------------------------------------------
sub nextrow
{
    my($self) = shift;
    my(%row);

    if ( $self->{dbproc}->FetchRow()) {
		undef %row;
		%row = $self->{dbproc}->DataHash();
	} else {
                print main::DEBUG "DBI_ODBC::nextrow - FAILED:".Win32::ODBC::Error()."\n" if $main::debug;
#%row = undef ;
 		undef %row;
	}

    print main::DEBUG "DBI_ODBC::nextrow=".join(',',%row)."\n" if $main::debug;

	return %row;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 disconnect()

Disconnects the current database connection

After this function is called there are no current database 
connection => no other function from this package can be 
called. 

This method is automatically called when the current object is
destroyed.

=cut
#------------------------------------------------------------------------------
sub disconnect
{
    my($self) = shift;

    print main::DEBUG "DBI_ODBC::destroy() \n" if $debug;
    $self->close();
    #$self->cache();
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 commit

Commits SQL statements executed since last commit

=cut
#------------------------------------------------------------------------------
sub commit
{
    my($self) = shift;

	# NOT IMPLEMENTED !
    WDB::Error::Fatal("WDB::DBI_ODBC::commit Not Implemented!\n");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 rollback

Rollsback SQL statements executed since last commit

=cut
#------------------------------------------------------------------------------
sub rollback
{
    my($self) = shift;

	# NOT IMPLEMENTED !
    WDB::Error::Fatal("WDB::DBI_ODBC::rollback Not Implemented!\n");
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 autocommit( on_or_off )

Enables automatic commit after each executed SQL statement.

If I<on_or_off> is true (non-zero) autocommit is enabled. If it is
false (zero) is is disabled.

=cut
#------------------------------------------------------------------------------
sub autocommit
{
    my($self)  = shift;
    my($onoff) = shift;

	# NOT IMPLEMENTED !
    WDB::Error::Fatal("WDB::DBI_ODBC::autocommit Not Implemented!\n");
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 dateformat( format )

Sets the default output format for dates.

Sets the order of the date parts month/day/year for entering
datetime data. Valid arguments are mdy, dmy, ymd, ydm, myd, dym.

=cut
#------------------------------------------------------------------------------
sub dateformat
{
    my($self)   = shift;
    my($format) = shift;
    
    print main::DEBUG "DBI_ODBC::dateformat( $format ) \n" if $main::debug;
    $self->do("ALTER SESSION SET NLS_DATE_FORMAT=\"$format\"");
}


1;#			    ----- EOF -----
