#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head1 NAME

Field - Definition of a FDF Field.

=head1 Description

This class defines a generic FDF field. There will be one Field subclass
per WDB field type.

The Field classes are also responsible for convertion between the WDB
types and the standard database types.

The Field classes are also responsible for producing the HTML to represent
them in the standard layouts. This will mostly be defined here in the Field
class, but can be overwritten in the Field subclasses.

=head1 Public Methods
=cut#--------------------------------------------------------------------------

package  WDB::Field;
use English;                    # Use long "english" variable names 

$debug = $main::debug;
#-------------------------------------------
# Define global variables for Field objects
#-------------------------------------------
%glob = ( MAXROWS   => 6,
	  LINEWIDTH => 65 );

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 new( %attr )

New creates a new Field object with the given attributes

=cut
#------------------------------------------------------------------------------
sub new 
{
    my($type) = shift;
    my($self) = {};
    my($fdf) = shift;
    my(%attr) = @_;
    bless ($self, $type);

    foreach $key ( keys %attr ) {
	$self->{$key} = $attr{$key};
    }
    $self->{'label'}  = $self->{FIELD}  if ( ! $self->{'label'} );
    $self->{'fdf'}    = $fdf;
    $self->{'glob'}   = \%glob;

    return $self;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 db_type()

Returns the database type of the field. 

=cut
#------------------------------------------------------------------------------
sub db_type
{
    my($self) = shift;

    return 'void';    # This should be overwritten by child classes !!!
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 default_oper()

Returns the default operand for queries

=cut
#------------------------------------------------------------------------------
sub default_oper
{
    my($self) = shift;

    return '=';    # This should be overwritten by child classes !!!
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 set_value( value )

Sets the value of the field.

=cut
#------------------------------------------------------------------------------
sub set_value
{
    my($self) = shift;

    $self->{value} = shift;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 set_db_value( value )

Sets the value of the field given the database value. 
(Converts using from_db)

=cut
#------------------------------------------------------------------------------
sub set_db_value
{
    my($self) = shift;
    my($val)  = shift;

    $self->{value} = $val;

    if ( $self->{'from_db'} ) {
	$self->{value} = WDB::Eval::eval($self->{FIELD}.':from_db', $self->{'from_db'});

        print main::DEBUG "set_db_value: from_db: $self->{'from_db'} = $self->{value}\n" if $main::debug;
    }
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 get_db_value

Returns the database value of the field.
(Converts using to_db)

=cut
#------------------------------------------------------------------------------
sub get_db_value
{
    my($self) = shift;
    my($val)  = $self->{value};

    print main::DEBUG "get_db_value: $self->{FIELD}\n" if $main::debug;
    if ( $self->{'to_db'} ) {
	$val = WDB::Eval::eval($self->{FIELD}.':to_db', $self->{'to_db'});

        print main::DEBUG "get_db_value: to_db $self->{'to_db'} = $val\n" if $main::debug;
    }

    return $val;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 set_query( value )

Sets the query value of the field.

TODO: UNUSED ?!?!? Use value instead....
=cut
#------------------------------------------------------------------------------
sub set_query
{
    my($self) = shift;

    $self->{query} = shift;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 set_oper( value )

Sets the query operand of the field.

=cut
#------------------------------------------------------------------------------
sub set_oper
{
    my($self) = shift;
    my($oper) = shift;

    $self->{oper} = ( $oper ? $oper : $self->default_oper );
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 verify( value )

Verifies that the given value is valid for this field.

Returns an error message. An empty string indicates success.

Each field also keeps the result of the last verify in {'errormsg'}

=cut
#------------------------------------------------------------------------------
sub verify
{
    my($self)     = shift;
    my($val)      = shift;
    my($errormsg) = '';

    $self->{'value'} = $val;

    if ($val eq '' ) {
	if ($self->{not_null}) {
	    $self->{'errormsg'} = sprintf($main::MSG{FIELD_OBL}, $self->{label});
	    return $self->{'errormsg'};
	} else {
	    $self->{'errormsg'} = '';
	    return $self->{'errormsg'};
	}
    };

    if ( $self->{'verify'} ) {
	$errormsg = WDB::Eval::eval($self->{FIELD}.':verify', $self->{'verify'});
    }

    $self->{'errormsg'} = $errormsg;
    return $errormsg;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 verify_query( value )

Verifies that the given value is valid for this field.

Returns an error message. An empty string indicates success.

Each field also keeps the result of the last verify in {'errormsg'}

=cut
#------------------------------------------------------------------------------
sub verify_query
{
    my($self)     = shift;
    my($val)      = shift;
    my($errormsg) = '';

    $self->{'value'} = $val;
    $self->{'errormsg'} = $errormsg;
    return $errormsg;
}


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 html( %attr )

Return the HTML for the current field in the given layout. This functions
uses the following named parameters :

    TYPE => 	Layout type : [query|text|label|unit|toggle]
    DOTS =>	Append dots and a colon to the label.
    LEN =>	Length of the input field

TODO: This method should be restructured, in to submethods, that can 
be overwritten individually from sub classes.

=cut
#------------------------------------------------------------------------------
sub html
{
    my($self) = shift;
    my(%attr) = @_;
    my($html) = '';
    my($value, $title, $length,$default);
    my($name);

    print main::DEBUG "WDB::Field::html(TYPE=$attr{TYPE}) for ". $self->{FIELD} ."\n" if $main::debug;

    WDB::Error::Fatal("No TYPE given in WDB::Field::html\n") if ( ! $attr{TYPE} );


    #------------------
    # TYPE => 'query'
    #------------------
    if ( $attr{TYPE} eq 'query' ) {

	#------------------
	# ENUM field ?
	#------------------
        if ( $self->{enum} ) {
	    print main::DEBUG "HTML for ENUM field: $self->{FIELD} (value=$self->{value})\n" if $main::debug;
            $html .= "<SELECT name=\"$self->{FIELD}\">";
            $default = $self->{fdf}->{cgi}->param($self->{FIELD});
            foreach ( split(',',$self->{enum}) ) {
                ( $value, $title ) = /^\s*(.*)=(.*)$/;
                #$html .= "<OPTION value=\"$value\"> $title\n";
                $html .= "<OPTION value=\"$value\"";
		if ( $value eq $default ) {
		    $html .= " SELECTED";
		}
		$html .= "> $title\n";
            }
            $html .= "</SELECT>";

	#------------------
	# Normal field !
	#------------------
	} else {
	    #----------------------------------
	    # Calculate length of input field
	    #----------------------------------
            $maxlength = $self->{'length'};
	    if ( $attr{LEN} ) {
		$length = $attr{LEN} ;
	    } elsif ( $self->{'length'} > ($self->{'glob'}->{LINEWIDTH} - $self->{'fdf'}->{'maxlabel'}) ) {
                $length = ($self->{'glob'}->{LINEWIDTH} - $self->{'fdf'}->{'maxlabel'}) ;
            } else {
                $length = $self->{'length'};
            }
	    #----------------------------------
	    # Default query string ?
	    #----------------------------------
	    print main::DEBUG "WDB::Field::html ". $self->{FIELD} ." default?" if $main::debug;
	    if ( $self->{fdf}->{cgi}->param($self->{FIELD}) ) {
                $default = $self->{fdf}->{cgi}->param($self->{FIELD});
		print main::DEBUG " from param: $default\n" if $main::debug;
            } elsif ( $self->{'default'} ) {
		$default = WDB::Eval::eval($self->{FIELD}.':default', $self->{'default'});
		print main::DEBUG " from FDF: $default\n" if $main::debug;
            } else {
                $default = "";
		print main::DEBUG " -none-\n" if $main::debug;
            }
	    #----------------------------------
	    # Default query string ?
	    #----------------------------------
            $html = "<INPUT name=\"$self->{FIELD}\" size=\"$length\" value=\"$default\">";
	}

    #------------------
    # TYPE => 'update'
    #------------------
    } elsif ( $attr{TYPE} eq 'update' or $attr{TYPE} eq 'insert' ) {
	my (@keys);
	my ($key);

	# Include keys in the name field for list updates ?
	if ( $attr{KEYS} ) {
	    $name = $self->{FIELD};
	    foreach $key ( @{$self->{fdf}->{keys}} ) {
	       push( @keys, &main::noblanks($self->{fdf}{field}{$key}{'value'}) );
	    }
	    $name .= '::' . join( '::', @keys );
	} elsif ( $attr{CHANGE} ) {
	    $name = 'change::'.$self->{FIELD};
	} else {
	    $name = $self->{FIELD};
	}

	#----------------------------------
	# Default string 
	#----------------------------------
	if ( $attr{CHANGE} ) {
	    $default = $self->{fdf}->{cgi}->param('change::'.$self->{FIELD}) ;
	} elsif ( $self->{value} ) {
	    $default = $self->{value};
	} elsif ( $self->{'ins_default'} and $attr{TYPE} eq 'insert' ) {
	    $default = WDB::Eval::eval($self->{FIELD}.':ins_default', $self->{'ins_default'});
	} else {
	    $default = '';
	}

	#------------------
	# ENUM field ?
	#------------------
        if ( $self->{enum} ) {
	    print main::DEBUG "HTML for ENUM filed: $self->{FIELD} (value=$default)\n" if $main::debug;
            $html .= "<SELECT name=\"$name\">";
            foreach ( split(',',$self->{enum}) ) {
                ( $value, $title ) = /^\s*(.*)=(.*)$/;
		next if ( $value eq '%' );
                $html .= "<OPTION value=\"$value\"";
		if ( $default eq $value ) {
		    $html .= " SELECTED";
		}
		$html .= "> $title\n";
            }
            $html .= "</SELECT>";

	#------------------
	# Normal field !
	#------------------
	} else {
	    #----------------------------------
	    # Calculate length of input field
	    #----------------------------------
            $maxlength = $self->{'length'};
            if ( $attr{LEN} ) {
		$length = $attr{LEN} ;
	    } elsif ( $self->{'length'} > ($self->{'glob'}->{LINEWIDTH} - $self->{'fdf'}->{'maxlabel'}) ) {
                $length = ($self->{'glob'}->{LINEWIDTH} - $self->{'fdf'}->{'maxlabel'}) ;
            } else {
                $length = $self->{'length'};
            }

	    #----------------------------------
	    # Text area ?
	    #----------------------------------
            if ( $self->{'textarea'} ) { 
		$rows = $self->{'length'}/$length;
		$rows = int($rows + 1);
		$rows = $self->{'glob'}->{MAXROWS} if $self->{'glob'}->{MAXROWS};
		$html = "<TEXTAREA COLS=$length ROWS=$rows WRAP=virtual"
		      . " NAME=\"$name\">"
		      . "$default</TEXTAREA>";
	    #----------------------------------
	    # Normal text input field
	    #----------------------------------
	    } else {
		$html = "<INPUT name=\"$name\" size=\"$length\" maxlength=\"$maxlength\" value=\"$default\">";
	    }
	}

    #-------------------------------------------------------------------------
    # TYPE => 'change_type'
    # Used together with CHANGE fields. (se above) and the 'change' action.
    #-------------------------------------------------------------------------
    } elsif ( $attr{TYPE} eq 'change_type' ) {
	my($name) = 'change_type::'.$self->{FIELD};
        my($default) = $self->{fdf}->{cgi}->param($name);

	print main::DEBUG "HTML for CHANGE_TYPE field: $self->{FIELD} (value=$self->{value})\n" if $main::debug;
	$html .= "<SELECT name=\"$name\">";
	foreach ( split(',',$self->{change}) ) {
	    ( $value, $title ) = /^\s*(.*)=(.*)$/;
	    next if ( $value eq '%' );
	    $html .= "<OPTION value=\"$value\"";
	    #if ( $self->{change_default} eq $value ) {
	    if ( $default eq $value or ( ! $default and $self->{change_default} eq $value ) ) {
		$html .= " SELECTED";
	    }
	    $html .= "> $title\n";
	}
	$html .= "</SELECT>";

    #------------------
    # TYPE => 'hidden'
    #------------------
    } elsif ( $attr{TYPE} eq 'hidden' ) {
	my (@keys);
	my ($key);
	# Include keys in the name field for list updates ?
	if ( $attr{KEYS} ) {
	    $name = $self->{FIELD};
	    foreach $key ( @{$self->{fdf}->{keys}} ) {
	       push( @keys, &main::noblanks($self->{fdf}{field}{$key}{'value'}) );
	    }
	    $name .= '::' . join( '::', @keys );
	} else {
	    $name = $self->{FIELD};
	}

	$html = "<INPUT type=hidden name=\"$name\" value=\"$self->{value}\">";

    #------------------
    # TYPE => 'text'
    #------------------
    } elsif ( $attr{TYPE} eq 'text' ) {
	$html = $self->_value;

    #------------------
    # TYPE => 'escaped'
    #------------------
    } elsif ( $attr{TYPE} eq 'escaped' ) {
	$html = &main::cgi_encode(&main::noblanks($self->{value}));

    #------------------
    # TYPE => 'label'
    #------------------
    } elsif ( $attr{TYPE} eq 'label' ) {
	$html = '<B>' . $self->{label} ;
	if ( $attr{DOTS} ) {
	    $html .= ('.' x ($self->{fdf}->{maxlabel} - length($self->{label}))). ':';
	}
	$html .= '</B>';

    #------------------
    # TYPE => 'unit'
    #------------------
    } elsif ( $attr{TYPE} eq 'unit' ) {
	print $self->{unitlabel} if $self->{unitlabel};

    #------------------
    # TYPE => 'toggle'   (Currently not fully implemeted....)
    #------------------
    } elsif ( $attr{TYPE} eq 'toggle' ) {
          $html  = "<INPUT type=checkbox ";
          $html .= "checked" if ! defined $self->{no_tab};
          $html .= " name=tab_$self->{FIELD}>";

    #------------------
    # TYPE => 'errormsg'
    #------------------
    } elsif ( $attr{TYPE} eq 'errormsg' ) {
	if ( $attr{CHANGE} ) {
	    if ( $self->{'change_errormsg'} ) {
		$html = "<FONT COLOR=\"red\">$self->{'change_errormsg'}</FONT>";
	    }
	} else {
	    if ( $self->{'errormsg'} ) {
		$html = "<FONT COLOR=\"red\">$self->{'errormsg'}</FONT>";
	    }
	}
    }

    return $html;
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#.PURPOSE	Compute output value of field (eval enums, pictures etc. ..)
#
#.REMARKS	This is temporary. Later this should be modularized a bit and
#		some of if should be moved to the Field_xxx classes.
#
#.RETURNS	New value;
#------------------------------------------------------------------------------
sub _value 
{
    my($self) = shift;
    my($val)  = $self->{value};
    my($newval);
    my($field) = $self->{fdf}->{field};

    print main::DEBUG "Value : $self->{FIELD} : $val " if $main::debug;

    if ( $val eq "NULL" 
	|| (defined $WDB::Eval::SPEC_NULL{$self->{'type'}} 
	    && $WDB::Eval::SPEC_NULL{$self->{'type'}} eq $val) ) {

	$val = $WDB::Eval::NULL_VALUE;
	return $val;
    } 

    #if ( $self->{'from_db'} ) {
    #	print main::DEBUG "(from_db)" if $debug;
    #	$val = WDB::Eval::eval($self->{FIELD}.':from_db', $self->{'from_db'});
    #}

    if ( $self->{'enum'} ) {
	( $void, $newval ) = $self->{'enum'} =~ /(^|[,\n\s])$val=([^,]+)/;
	
	print main::DEBUG "(enum) $self->{'FIELD'} : $self->{'enum'}\n" if $debug;
	print main::DEBUG "(enum) $self->{'FIELD'} ($val) = $newval\n" if $debug;

	$val = $newval if ( $newval );
    }

    if ( $self->{'url'} ) {
	print main::DEBUG "(url)" if $debug;
	$url = WDB::Eval::eval($self->{FIELD}.':url', $self->{'url'});
        $url = &main::noblanks($url);
	$val = "<a href=\"$url\">$val</a>" if $url;
    }
    
    print main::DEBUG "= $val\n" if $debug;
    return  $val;
}

1; #			     ----- The End ----- 
