#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head1 action_del_released

The 'del_released' action for WDB.

=head1 Description

Deletes released data from the DARWIN datase. This is an action developed 
specificaly for the DARWIN project. It requires the FDF to define two extra 
form attributes (+ one optional).

RELEASE_TABLE = name of table containing released data.
RELEASE_UPDATE= update action to perform on the row in the 'working' table 
		to indicate that the data has been released. If left empty 
		or undefined no update is performed.
RELEASE_SUB   = Defines 'sub-tables' to also release data from. This is a 
		complex data structure with the following syntax :

		TDB...


This module is automatically loaded by WDB when the 'del_released' action is
used.

=head1 Functions
=cut#--------------------------------------------------------------------------


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 action_del_released

Deletes DARWIN data.

=cut
#------------------------------------------------------------------------------
sub action_del_released
{
    my($cgi,$fdf) = @_;
    my($dbi,@keys,$key);
    my($val);

    print main::DEBUG "*** action_del_released ***\n" if $main::debug;

    #------------------------
    # VERIFY
    #------------------------
    if ( ! $fdf->verify($cgi) ) {
	WDB::Error::Fatal("Unexpected error. Verify failed in action_del_released.");
    } else {
	$WDB::Eval::ERRORMSG = '';
    }

    #------------------------
    # LOAD DBI
    #------------------------
    $dbi = load_dbi();

    #------------------------------------
    # COMPOSE SQL AND RELEASE TABLE
    #------------------------------------
    my($table) = $fdf->{'TABLE'} =~ /([^,]+).*/; 

    foreach $f ( @{$fdf->{fields}} ) {
	#------------------------
	# Key field ?
	#------------------------
        if ( defined $f->{'key'} ) {
	    print main::DEBUG "Delete Released: $f->{FIELD} - key" if $main::debug;

	    $val = $f->get_db_value();
            if ( ! defined $val || $val eq '' ) {
                WDB::Error::Fatal("Fatal error: missing key data.");
            } else {
		my($col) = $f->{column} =~ /.*?\.?([^\.]+)$/;
	        print main::DEBUG "($col)\n" if $debug;
		push( @where, eval('$dbi->query_' . $f->db_type 
			       . '($col,$f->{oper},$f->{value})') );
            }

        }
    }

    my($where) = join(' AND ',@where);
    do_del_released( $table, $fdf->{RELEASE_TABLE}, $where, 0, $fdf->{RELEASE_SUB} ); 

    #----------------------------------
    # PRINT RESULT.
    #----------------------------------

    $fdf->layout( ( $cgi->param('wdb_ldf') ? $cgi->param('wdb_ldf') : 'del_released') );
    $fdf->print_html;
    #$dbi->close();
}

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=head2 do_del_released

Deletes one or more rows from a table, and the cooresponding released data 
in a parallel table.

arguments:
    $table		Table to delete from 
    $rel_table		Table containing released data (also deleted if it exits)
    $where		SQL 'where' clause identifying data in the table(s)
    $keys		Key fields in the 'from' table. (used in update)
    $release_sub 	Special structure defining sub-tables to release.
=cut
#------------------------------------------------------------------------------
sub do_del_released
{
    my($table, $rel_table, $where, $keys, $release_sub ) = @_;
    my($sql_stmt,%row);
    my($dbi) = load_dbi();
    my(@keys) = split(':', $keys);

    print main::DEBUG "\n********** do_del_released **********\n"
		    . "table......: $table\n"
		    . "rel_table .: $rel_table\n"
		    . "where......: $where\n"
		    . "keys.......: $keys\n"
		    . "release_sub: $release_sub\n"
		    if $main::debug;

    #----------------------------------
    # RELEASE SUB-TABLES (IF ANY)
    #----------------------------------
    if ( $release_sub ) {
	$dbi->open("select * from $table where $where");

	while ( %row = $dbi->nextrow ) {
	    my($where) = $where;

	    #---------------------------------------------------
	    # If keys are defined, construct new where clause.  
	    #---------------------------------------------------
	    if ($keys) {
		my(@kw,$key);
		foreach $key ( @keys ) {
		    push(@kw, $key . "='$row{$key}'" );
		}
		$where = join(' AND ',@kw);
	    }

	    my($sub);
	    foreach $sub ( split( ';', $release_sub ) ) {
		my($table, $rel_table, $where, $update, $keys ) = split(',',$sub);
		#-------------------------
		# Modify where clause....
		#-------------------------
		print main::DEBUG "Modifying where: $where\n" if $main::debug;
		while ( $where =~ /\$(\w+)/ ) {
		    $new_where = $` . $row{$1};
		    print main::DEBUG "  $1 = $row{$1}\n" if $main::debug;
		    $where = $';
		}
		$where = $new_where . $where;

		#-------------------------
		# Release sub-table data
		#-------------------------
		do_del_released( $table, $rel_table, $where, $keys, '' ); 
	    }
	}
    }

    $dbi->do("delete from $table where $where");
    $dbi->try("delete from $rel_table where $where") if $rel_table;
}

1;
